# Copyright 2013-2017 by Luc Saffre.
# License: BSD, see LICENSE for more details.

"""
Defines an extended TestCase whith methods to launch a subprocess.

"""
from __future__ import unicode_literals

import six
import os
from os.path import join
import unittest
import glob
from fnmatch import fnmatch
import sys
from setuptools import find_packages
# from unipath import Path
import subprocess

# from atelier import SETUP_INFO
from atelier.utils import SubProcessParent

# ROOTDIR = Path(__file__).parent.parent
DOCTEST_CMD = "atelier.doctest_utf8"


def interpreter_args():
    # if 'coverage' in sys.modules:
    #     # raise Exception('20160119')
    #     return ['coverage', 'run']
    return [sys.executable]


class DocTestCase(unittest.FunctionTestCase, SubProcessParent):
    # internally used by make_docs_suite
    
    def build_environment(self):
        env = super(DocTestCase, self).build_environment()
        env.pop('PYTHONPATH', None)  # fixes #1296
        return env
    
    def __init__(self, filename):
        def func():
            args = [sys.executable]
            args += ["-m"]
            args += [DOCTEST_CMD]
            args += [filename]
            self.run_subprocess(args)
        func.__name__ = filename
        super(DocTestCase, self).__init__(func)
    

def make_docs_suite(docs_root, include="*.rst", exclude=None):
    """Discover the doc files in specified directory docs_root and below
    and return a test suite which tests them all, each one in a
    separate subprocess.

    `include` is a filename pattern of the files to include. Default
    is `'*.rst'`.

    `exclude` is an optional filename pattern of the files to
    exclude. Default is None.

    """
    suite = unittest.TestSuite()
    for root, dirs, files in os.walk(docs_root):
        for file in files:
            fn = join(root, file)
            if fnmatch(fn, include):
                if exclude and fnmatch(fn, exclude):
                    continue
                suite.addTest(DocTestCase(fn))
    return suite




class TestCase(unittest.TestCase, SubProcessParent):
    "Deserves a docstring"
    project_root = NotImplementedError
    # maxDiff = None

    def run_packages_test(self, declared_packages):
        """
        Checks whether the `packages` parameter to setup seems correct.
        """
        found_packages = find_packages()
        # if tests exists, remove it:
        if 'tests' in found_packages:
            found_packages.remove('tests')
        found_packages.sort()
        declared_packages.sort()
        self.assertEqual(found_packages, declared_packages)

    def run_simple_doctests(self, filenames, **kw):  # env.simple_doctests
        """
        run doctest of given file in a subprocess
        """
        for ln in filenames.splitlines():
            ln = ln.strip()
            if ln and not ln.startswith('#'):
                ok = False
                for fn in glob.glob(ln):
                    ok = True
                    args = interpreter_args()
                    args += ["-m"]
                    # args += ["doctest"]
                    args += ["atelier.doctest_utf8"]
                    args += [fn]
                    self.run_subprocess(args, **kw)
                if not ok:
                    self.fail("no files matching %s" % ln)

    def run_unittest(self, filename, **kw):
        """
        run unittest of given file in a subprocess
        """
        args = interpreter_args()
        #~ args += ["-Wall"]
        args += ["-m"]
        args += ["unittest"]
        #~ args += ["--buffer"]
        args += [filename]
        self.run_subprocess(args, **kw)

    # ~ def run_simple_doctests(self,filename,**kw): # env.simple_doctests
        #~ doctest.testfile(os.path.abspath(filename),
            #~ encoding='utf-8',
            #~ module_relative=False)


# class BaseTestCase(TestCase):
#     project_root = ROOTDIR
    

# class BasicTests(BaseTestCase):

#     def test_01(self):
#         self.assertEqual(1+1, 2)

#     def test_utils(self):
#         self.run_simple_doctests('atelier/utils.py')

#     def test_rstgen(self):
#         self.run_simple_doctests('atelier/rstgen.py')


# class PackagesTests(BaseTestCase):
#     def test_packages(self):
#         self.run_packages_test(SETUP_INFO['packages'])


# class SphinxTests(BaseTestCase):
#     def test_sphinxconf(self):
#         self.run_simple_doctests('atelier/sphinxconf/__init__.py')

#     def test_base(self):
#         self.run_simple_doctests('atelier/sphinxconf/base.py')

#     def test_sigal(self):
#         self.run_simple_doctests('atelier/sphinxconf/sigal_image.py')





# -*- coding: UTF-8 -*-
# Copyright 2016-2017 by Luc Saffre & Hamza Khchine.
# License: BSD, see LICENSE for more details.
"""An extensible library of `invoke
<http://docs.pyinvoke.org/en/latest/index.html>`__ tasks.
See :doc:`/invlib`.

.. autosummary::
   :toctree:

   dummy

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os

from invoke import Collection
from unipath import Path

import atelier

def setup_from_tasks(
        self, globals_dict, main_package=None,
        settings_module_name=None, **kwargs):
    """This is the function you must call from your :xfile:`tasks.py` file
    in order to activate the tasks defined by atelier. 

    """
    if '__file__' not in globals_dict:
        raise Exception(
            "No '__file__' in %r. "
            "First parameter to must be `globals()`" % globals_dict)

    tasks = Path(globals_dict['__file__'])
    if not tasks.exists():
        raise Exception("No such file: %s" % tasks)
    root_dir = tasks.parent.absolute()
    configs = {
        'root_dir': root_dir,
        'main_package': main_package,
        'locale_dir': None,
        'help_texts_source': None,
        'help_texts_module': None,
        'tolerate_sphinx_warnings': False,
        'cleanable_files': [],
        'revision_control_system': None,
        'apidoc_exclude_pathnames': [],
        'project_name': tasks.parent.absolute().name,
        'editor_command': None,
        'coverage_command': 'setup.py test',
        'languages': None,
        'blog_root': root_dir.child('docs'),
        'long_date_format': "%Y%m%d (%A, %d %B %Y)",
        'sdist_dir': root_dir.child('dist'),
        'pypi_dir': root_dir.child('.pypi_cache'),
    }

    if settings_module_name is not None:
        os.environ['DJANGO_SETTINGS_MODULE'] = settings_module_name
        from django.conf import settings
        self.configure({
            'languages': [lng.name for lng in settings.SITE.languages]})

    configs.setdefault(
        'build_dir_name', '.build')  # but ablog needs '_build'
    configs.setdefault('use_dirhtml', False)

    # # The following import will populate the projects
    from atelier.projects import get_project_info_tasks
    prj = get_project_info_tasks(root_dir)
    prj.load_tasks()

    # we cannot store current_project using configure() because it
    # cannot be pickled. And we don't need to store it there, it is
    # not a configuration value but just a global internal variable.
    atelier.current_project = prj
    
    self.main_package = main_package
    
    self.configure({'doc_trees': prj.doc_trees})
    self.configure(configs)
    if kwargs:
        self.configure(kwargs)
    # return _globals_dict


class MyCollection(Collection):
    
    def setup_from_tasks(self, *args, **kwargs):
        return setup_from_tasks(self, *args, **kwargs)

    @classmethod
    def from_module(cls, *args, **kwargs):
        """
        A hack needed to make it work as long as invoke does not yet
        support subclassing Collection
        (https://github.com/pyinvoke/invoke/pull/342)
        """
        
        ns = super(MyCollection, cls).from_module(*args, **kwargs)

        if ns.__class__ != cls:
            from functools import partial
            ns.setup_from_tasks = partial(setup_from_tasks, ns)
        return ns




# Copyright 2013-2017 by Luc Saffre.
# License: BSD, see LICENSE for more details.

# ~ This module has no docstring because it is to be execfile'd
# ~ from `setup.py`, `atelier/__init__.py` and possibly some external
# ~ tools, too.

# Explicitly install `importlib` under Python 2.6. Thanks to
# http://stackoverflow.com/questions/9418064
install_requires = ['Sphinx', 'invoke', 'argh', 'six',
                    'future', 'Babel', 'unipath',
                    'python_dateutil', 'gitpython', 'pytest-cov']
try:
    import importlib
except ImportError:
    install_requires.append('importlib')

SETUP_INFO = dict(
    name='atelier',
    version='1.0.8',
    install_requires=install_requires,
    scripts=['scripts/per_project'],
    description="A collection of tools for software artists",
    license='Free BSD',
    test_suite='tests',
    author='Luc Saffre',
    author_email='luc.saffre@gmail.com',
    url="http://atelier.lino-framework.org",
    long_description="""\

.. image:: https://readthedocs.org/projects/atelier/badge/?version=latest
   :target: http://atelier.readthedocs.io/en/latest/?badge=latest
.. image:: https://coveralls.io/repos/github/lino-framework/atelier/badge.svg?branch=master
   :target: https://coveralls.io/github/lino-framework/atelier?branch=master
.. image:: https://travis-ci.org/lino-framework/atelier.svg?branch=master
   :target: https://travis-ci.org/lino-framework/atelier?branch=master
.. image:: https://img.shields.io/pypi/v/atelier.svg
   :target: https://pypi.python.org/pypi/atelier/
.. image:: https://img.shields.io/pypi/l/atelier.svg
   :target: https://pypi.python.org/pypi/atelier/

`atelier` is a collection of tools for managing and maintaining
Python software projects.

It contains:

- some general Python utilities
  (`atelier.utils <http://atelier.lino-framework.org/api/atelier.utils.html>`_)
- a library for generating reStructuredText from Python
  (`atelier.rstgen <http://atelier.lino-framework.org/api/atelier.rstgen.html>`_)
- some Sphinx extensions
  (`atelier.sphinxconf <http://atelier.lino-framework.org/api/atelier.sphinxconf.html>`_)
- a library of invoke commands
  (`atelier.invlib <http://atelier.lino-framework.org/api/atelier.invlib.html>`_)
- a minimalistic project management
  (`atelier.projects <http://atelier.lino-framework.org/api/atelier.projects.html>`_)

The central project homepage is http://atelier.lino-framework.org

""",
    classifiers="""\
Programming Language :: Python
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Framework :: Sphinx :: Extension
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
License :: OSI Approved :: BSD License
Natural Language :: English
Operating System :: OS Independent""".splitlines())

SETUP_INFO.update(packages=[n for n in """
atelier
atelier.sphinxconf
atelier.invlib
""".splitlines() if n])

SETUP_INFO.update(package_data=dict())


def add_package_data(package, *patterns):
    l = SETUP_INFO['package_data'].setdefault(package, [])
    l.extend(patterns)
    return l


add_package_data('atelier.sphinxconf', '*.html')



# -*- coding: UTF-8 -*-
# Copyright 2016-2017 by Luc Saffre & Hamza Khchine.
# License: BSD, see LICENSE for more details.

from atelier.invlib import MyCollection
from . import tasks

ns = MyCollection.from_module(tasks)
configs = dict(
    demo_projects=[],
    prep_command="python manage.py prep --noinput --traceback")
ns.configure(configs)




# -*- coding: utf-8 -*-
# Copyright 2011-2016 by Luc Saffre.
# License: BSD, see LICENSE for more details.

"""This Sphinx extension defines the :rst:dir:`blogger_year` and
:rst:dir:`blogger_index` directives.

Usage: add the following to your `conf.py`::

  extensions += ['atelier.sphinxconf.blog']

And usually this file structure:

- docs/blog/index.rst --> contains a main_blogindex directive (hidden toctree)
    
Individual blog entries, including yearly directories and
`index.rst` files, are automatically created by :cmd:`fab blog`,
leading to a file structure like this:
    
- docs/blog/2013/index.rst --> contains a :rst:dir:`blogger_year` directive
- docs/blog/2013/0107.rst --> a blog entry
- docs/blog/2010/0107.rst
   
Thanks to

- `Creating reStructuredText Directives
  <http://docutils.sourceforge.net/docs/howto/rst-directives.html>`_

"""
from past.builtins import cmp
from builtins import str
from builtins import map
from builtins import range
from builtins import object

import os
import calendar
import datetime

import jinja2

from babel.dates import format_date

from .insert_input import InsertInputDirective
from atelier.rstgen import toctree


def monthname(n, language):
    """
    Return the monthname for month # n in specified language.
    """
    d = datetime.date(2013, n, 1)
    return format_date(d, "MMMM", locale=language)


templates = dict()
templates['calendar.rst'] = """
====
{{year}}
====

{{intro}}

.. |br| raw:: html

   <br />   

.. |sp| raw:: html

   <span style="color:white;">00</span>

{{calendar}}


"""

JINJA_ENV = jinja2.Environment(
    #~ extensions=['jinja2.ext.i18n'],
    loader=jinja2.DictLoader(templates))


class BloggerYear(object):

    """A :class:`BloggerYear` instance is created for each `blogger_year`
    directive.

    """

    def __init__(self, env):
        """
        :docname: the name of the document containing the `main_blogindex` directive
        :starting_year: all years before this year will be pruned
        """

        blogname, year, index = env.docname.rsplit('/', 3)
        if index != 'index':
            raise Exception(
                "Allowed only in /<blogname>/<year>/index.rst files")
        self.blogname = blogname
        self.year = int(year)

        #~ print "20130113 Year.__init__", blogname, self.year
        #~ self.blogname = blogname
        self.days = []
        self.dates = set()
        #~ self.years = set()
        #~ self.starting_year = int(starting_year)
        top = os.path.dirname(env.doc2path(env.docname))
        #~ print top
        for (dirpath, dirnames, filenames) in os.walk(top):
            del dirnames[:]  # don't descend another level
            #~ unused, year = dirpath.rsplit(os.path.sep,2)
            #~ year = int(year)
            #~ assert year in self.years
            d = None
            for fn in sorted(filenames):
                if fn.endswith('.rst'):
                    docname = fn[:-4]
                    if docname == "index":
                        continue
                    if len(fn) == 8:
                        d = docname_to_day(self.year, docname)
                        self.days.append(d)
                        self.dates.add(d.date)
                        #~ self.years.add(s)
                    elif d is not None:
                        d.docnames.append(docname)

        #~ self.years = sorted(self.years)
        if not hasattr(env, 'blog_instances'):
            env.blog_instances = dict()
        years = env.blog_instances.setdefault(blogname, dict())
        years[self.year] = self


def docname(y):
    assert isinstance(y, BloggerYear)
    return "/" + y.blogname + "/" + str(y.year) + "/index"


def navigator(years, current):
    chunks = []
    for y in years:
        if y == current:
            chunks.append(str(y.year))
        else:
            chunks.append(
                ":doc:`{0} <{1}>`".format(y.year, docname(y)))
    old = ' '.join(chunks)
    return "\n\n{0}\n\n".format(old)


def get_blogger_years(env, blogname):
    blog_instances = getattr(env, 'blog_instances', dict())
    blog = blog_instances.get(blogname)
    if blog is None:
        return []
    years = list(blog.values())

    years.sort(key=lambda f: f.year)
    return years


class MainBlogIndexDirective(InsertInputDirective):
    """
    Directive to insert a blog master archive page toctree
    """
    #~ required_arguments = 1
    #~ allow_titles = True
    raw_insert = True

    def get_rst(self):
        #~ print 'MainBlogIndexDirective.get_rst()'
        env = self.state.document.settings.env
        blogname, index = env.docname.rsplit('/', 2)
        if index != 'index':
            raise Exception("Allowed only inside index.rst file")
        text = ''
        years = get_blogger_years(env, blogname)
    
        hidden = []
        visible = []

        # for y in years:
            # text += "\n    {0}/index".format(blogger_year.year)
            # visible.append(str(blogger_year.year) + "/index")
        if len(years) > 1:
            hidden = years[:-1]
            visible = years[-1:]
        else:
            visible = years

        text += navigator(years, None)

        text += '\n'.join(self.content)
        if len(years) == 0:
            text += "\n\nNo blogger years found.\n"
        else:
            children = list(map(docname, years))
            text += toctree(*children, hidden=True)

        # if len(hidden):
        #     children = map(docname, hidden)
        #     text += toctree(*children, hidden=True)
        # if len(visible):
        #     children = map(docname, visible)
        #     text += toctree(*children, maxdepth=2)
        # text += "\n"
        #~ print text
        return text


class YearBlogIndexDirective(InsertInputDirective):

    """
    Directive to insert a year's calendar
    """
    #~ required_arguments = 1
    #~ allow_titles = True
    raw_insert = True

    def get_rst(self):

        env = self.state.document.settings.env
        today = datetime.date.today()

        blogger_year = BloggerYear(env)
        years = get_blogger_years(env, blogger_year.blogname)
        tpl = JINJA_ENV.get_template('calendar.rst')

        intro = navigator(years, blogger_year)
        intro += '\n'.join(self.content)

        text = ''
        cal = calendar.Calendar()
        for month in range(1, 13):

            text += """
            
.. |M%02d| replace::  **%s**""" % (month, monthname(month, self.language))

            weeknum = None
            #~ text += "\n  |br| Mo Tu We Th Fr Sa Su "
            for day in cal.itermonthdates(blogger_year.year, month):
                iso_year, iso_week, iso_day = day.isocalendar()
                if iso_week != weeknum:
                    text += "\n  |br|"
                    weeknum = iso_week
                if day.month == month:
                    label = "%02d" % day.day
                    docname = "%02d%02d" % (day.month, day.day)
                    # if blogger_year.year == iso_year and day in blogger_year.days:
                    if day in blogger_year.dates:
                        text += " :doc:`%s <%s>` " % (label, docname)
                    elif day > today:
                        text += ' |sp| '
                    else:
                        text += ' ' + label + ' '
                else:
                    text += ' |sp| '

        text += """
        
===== ===== =====
|M01| |M02| |M03|
|M04| |M05| |M06|
|M07| |M08| |M09|
|M10| |M11| |M12|
===== ===== =====
        
        """

        text += """

.. rubric:: {0}

.. toctree::
    :maxdepth: 2
    
""".format("All entries:")

        for day in blogger_year.days:
            for docname in day.docnames:
                text += ("\n    " + docname)
    #         text += """
    # %02d%02d""" % (day.month, day.day)

        return tpl.render(
            calendar=text,
            intro=intro,
            year=blogger_year.year)
            # days=blogger_year.days)

class BloggerDay(object):
    def __init__(self, docname, *args, **kwargs):
        self.docnames = [docname]
        self.date = datetime.date(*args, **kwargs)

def docname_to_day(year, s):
    #~ print fn
    month = int(s[:2])
    day = int(s[2:])
    return BloggerDay(s, year, month, day)


#~ class ChangedDirective(InsertInputDirective):

    #~ def get_rst(self):
        #~ env = self.state.document.settings.env
        #~ blogname, year, monthday = env.docname.rsplit('/',3)
        # ~ # raise Exception("Allowed only in blog entries")

        #~ year = int(year)
        #~ day = docname_to_day(year,monthname)

        #~ if not hasattr(env,'changed_items'):
            #~ env.changed_items = dict()
        #~ env.changed_items
        #~ for item in self.content:
            #~ entries = env.changed_items.setdefault(item,dict())
            #~ entries.setdefault(env.docname)


def setup(app):
    #~ app.add_node(blogindex)
    #~ app.add_node(blogindex,html=(visit_blogindex,depart_blogindex))
    #~ app.add_directive('changed', ChangedDirective)
    # app.add_config_value('blog_instances', dict(), '')
    app.add_directive('blogger_year', YearBlogIndexDirective)
    app.add_directive('blogger_index', MainBlogIndexDirective)



# -*- coding: UTF-8 -*-
# Copyright 2013-2016 by Luc Saffre.
# License: BSD, see LICENSE for more details.

"""A wrapper for Python's doctest.

Because the command-line interface of `python -m doctest` has no way
to specify an encoding of a (non-.py) input file.

Code originally copied from Python 2.7 doctest.py

About the :message:`ValueError: Attempted relative import in
non-package` problem: see `Doctest and relative imports
<http://stackoverflow.com/questions/4047227/doctest-and-relative-imports>`_
where Jason Scheirer advises:

    I would get rid of the relative imports, many Python style guides
    strongly discourage them, most forbid them.

"""
from __future__ import print_function

import os
import sys
import doctest


def _test():

    testfiles = [arg for arg in sys.argv[1:] if arg and arg[0] != '-']
    if len(testfiles) == 0:
        name = os.path.basename(sys.argv[0])
        if '__loader__' in globals():          # python -m
            name, _ = os.path.splitext(name)
        print("usage: {0} [-v] file ...".format(name))
        return 2

    for filename in testfiles:
        if filename.endswith(".py"):
            # It is a module -- insert its dir into sys.path and try to
            # import it. If it is part of a package, that possibly
            # won't work because of package imports.
            dirname, filename = os.path.split(filename)
            sys.path.insert(0, dirname)
            m = __import__(filename[:-3])
            del sys.path[0]
            failures, _ = doctest.testmod(m)
            #~ raise Exception("20131022 tested %s" % m)
        else:
            failures, _ = doctest.testfile(
                os.path.abspath(filename),
                optionflags=doctest.REPORT_ONLY_FIRST_FAILURE,
                encoding='utf-8',
                module_relative=False)

        if failures:
            return 1
    return 0


if __name__ == "__main__":
    sys.exit(_test())



# -*- coding: utf-8 -*-
# Copyright 2011-2016 by Luc Saffre.
# License: BSD, see LICENSE for more details.

"""Sphinx setup used to build most of the documentation trees
mainained by myself.


.. toctree::

.. autosummary::
   :toctree:

   insert_input
   refstothis
   sigal_image
   complex_tables
   blog
   base
   dirtables
   interproject



"""

from __future__ import print_function
from __future__ import unicode_literals
#from builtins import str

"""Note: the `import unicode_literals` caused the following::

    Traceback (most recent call last):
      File "/home/luc/pythonenvs/py27/local/lib/python2.7/site-packages/sphinx/cmdline.py", line 245, in main
        warningiserror, tags, verbosity, parallel)
      File "/home/luc/pythonenvs/py27/local/lib/python2.7/site-packages/sphinx/application.py", line 122, in __init__
        self.config.setup(self)
      File "/home/luc/hgwork/atelier/atelier/sphinxconf/__init__.py", line 654, in setup
        indextemplate='pair: %s; management command')
      File "/home/luc/pythonenvs/py27/local/lib/python2.7/site-packages/sphinx/application.py", line 503, in add_object_type
        'doc_field_types': doc_field_types})
    TypeError: type() argument 1 must be string, not unicode
    
I solved this by a manual patch in line 308 of 
:file:`sphinx/application.py`::
    
    def import_object(self, objname, source=None):
        objname = str(objname)  # LS 20140108 accept unicode strings
        # needed when calling from Python 2.7 with
        # `from __future__ import unicode_literals`
        try:
            module, name = objname.rsplit('.', 1)
        except ValueError, err:

"""

import logging
logger = logging.getLogger(__name__)

import os
import sys

from unipath import Path

def configure(globals_dict):
    """Adds to your `conf.py` an arbitrary series of things that all my
    Sphinx docs configuration files have in common.

    To be called from inside the Sphinx `conf.py` as follows::

      from djangosite.utils.sphinxconf import configure
      configure(globals())

    Incomplete list of `conf.py` settings that will be set:

    - `extensions`
    - `templates_path`
    - `html_static_path`
    - master_doc='index'
    - source_suffix='.rst'
    - primary_domain='py'
    - pygments_style='sphinx'


    """
    filename = globals_dict['__file__']
    sys.path.append(Path(filename).parent.absolute())

    globals_dict.update(extensions=[
        'sphinx.ext.autodoc',
        'sphinx.ext.autosummary',
        'sphinx.ext.inheritance_diagram',
        'sphinx.ext.todo',
        'sphinx.ext.extlinks',
        'sphinx.ext.graphviz',
        'sphinx.ext.intersphinx',
        # no i18n, no discovery, only one entry per doc,
        # 'sphinxcontrib.newsfeed',
        #~ 'sphinx.ext.doctest',
        'atelier.sphinxconf.base',
        'atelier.sphinxconf.dirtables',
        'atelier.sphinxconf.refstothis',
        'atelier.sphinxconf.insert_input',
    ])

    # default config for autosummary:
    globals_dict.update(autosummary_generate=True)
    globals_dict.update(autodoc_default_flags=['members'])

    mydir = Path(__file__).parent.absolute()
    globals_dict.update(templates_path=['.templates', mydir])

    # globals_dict.update(html_static_path=['.static'])

    # some settings i use in all projects:

    globals_dict.update(master_doc='index')
    globals_dict.update(source_suffix='.rst')
    globals_dict.update(primary_domain='py')
    globals_dict.update(pygments_style='sphinx')

    globals_dict.update(autodoc_member_order='bysource')
    globals_dict.update(autodoc_default_flags=['show-inheritance', 'members'])

    my_font_family = "Swiss, Helvetica, 'Liberation Sans'"
    globals_dict.update(html_theme_options={
        "font_family": my_font_family,
        "head_font_family": my_font_family,
    })


    # globals_dict.update(
    #     blogref_format="http://www.lino-framework.org/blog/%Y/%m%d.html")


def version2rst(self, m):
    """
    used in docs/released/index.rst
    """
    v = m.__version__
    if v.endswith('+'):
        v = v[:-1]
        print("The current stable release is :doc:`%s`." % v)
        print("We are working on a future version in the code repository.")
    elif v.endswith('pre'):
        print("We're currently working on :doc:`%s`." % v[:-3])
    else:
        print("The current stable release is :doc:`%s`." % v)
        #~ print("We're currently working on :doc:`coming`.")


from sphinx.jinja2glue import BuiltinTemplateLoader


class DjangoTemplateBridge(BuiltinTemplateLoader):

    """The :meth:`configure` method installs this as `template_bridge
    <http://sphinx-doc.org/config.html#confval-template_bridge>`_ for
    Sphinx.  It causes a template variable ``settings`` to be added
    the Sphinx template context. This cannot be done using
    `html_context
    <http://sphinx-doc.org/config.html#confval-html_context>`_ because
    Django settings are not pickleable.

    """

    def render(self, template, context):
        from django.conf import settings
        context.update(settings=settings)
        return super(DjangoTemplateBridge, self).render(template, context)

    def render_string(self, source, context):
        from django.conf import settings
        context.update(settings=settings)
        return super(DjangoTemplateBridge, self).render_string(source, context)





# -*- coding: UTF-8 -*-
# Copyright 2013-2016 by Luc Saffre.
# License: BSD, see LICENSE for more details.

"""A library for `fabric <http://docs.fabfile.org>`__ with tasks I use
to manage my Python projects.

NOTE: This module is deprecated. Use :mod:`atelier.invlib` instead.

.. contents::
  :local:

.. _fab_commands:

``fab`` commands
================

Documenting
-----------

.. command:: fab blog

    Edit today's blog entry, create an empty file if it doesn't yet exist.

.. command:: fab cd

    Output a reStructuredText formatted list of all commits in all
    projects today.

.. command:: fab bd

    Converted to :cmd:`inv bd`.

.. command:: fab pd

    Converted to :cmd:`inv pd`.

.. command:: fab clean

    Converted to :cmd:`inv clean`.

.. command:: fab readme

    Converted to :cmd:`inv readme`.

.. command:: fab api

    No longer exists because we now use autosummary instead of
    sphinx-apidoc.

    Generate `.rst` files below `docs/api` by running `sphinx-apidoc
    <http://sphinx-doc.org/invocation.html#invocation-of-sphinx-apidoc>`_.

    This is no longer used by most of my projects, at least those
    which I converted to `sphinx.ext.autosummary`.


.. command:: fab docs

    Has been replaced by :cmd:`inv bd`.

.. command:: fab pub

    Has been replaced by :cmd:`inv pd`.


Internationalization
--------------------

.. command:: fab mm

    Converted to :cmd:`inv mm`.

Deploy
------

.. command:: fab release

    Converted to :cmd:`inv release`.

.. command:: fab sdist

    Converted to :cmd:`inv sdist`.

.. command:: fab ci

    Converted to :cmd:`inv ci`.

.. command:: fab reg

    Converted to :cmd:`inv reg`.




Testing
-------

.. command:: fab initdb

    Converted to :cmd:`inv initdb`.

.. command:: fab test

    Converted to :cmd:`inv test`.

.. command:: fab test_sdist

    Converted to :cmd:`inv test_sdist`.

Miscellaneous
-------------

.. command:: fab summary

    Converted to :cmd:`inv ls`.


Installation
============

To be used by creating a :file:`fabfile.py` in your project's root
directory with at least the following two lines::

  from atelier.fablib import *
  setup_from_fabfile(globals())

See :func:`setup_from_fabfile` for more information.



Configuration files
===================

.. xfile:: fabfile.py

In your :xfile:`fabfile.py` file you can specify project-specific
configuration settings.  Example content::

  from atelier.fablib import *
  setup_from_fabfile(globals(), "foobar")
  env.languages = "de fr et nl".split()
  env.tolerate_sphinx_warnings = True
  add_demo_project('foobar.demo')

.. xfile:: .fabricrc

To specify certain default preferences for all your projects, you can
create a file named :xfile:`.fabricrc` in your home directory with
content like this::

    user = luc
    blogger_project = lino
    docs_rsync_dest = luc@example.org:~/public_html/%s
    docs_rsync_dest = luc@example.org:~/public_html/{prj}_{docs}
    sdist_dir = /home/luc/projects/lino/docs/dl
    temp_dir = /home/luc/tmp


Project settings
================

`fabric <http://docs.fabfile.org>`__ works with a global "environment"
object named ``env``.

The following section documents the possible attributes of this object
as used by :mod:`atelier.fablib`.

You usually define these in your :xfile:`fabfile.py`.  For some of
them (those who are simple strings) you can define user-specific
default values in a :xfile:`.fabricrc` file.


.. class:: env

  .. attribute:: locale_dir

    The name of the directory where `fab mm` et al should write their
    catalog files.

  .. attribute:: sdist_dir


  .. attribute:: editor_command

    A string with the command name of your text editor. Example::

      editor_command = "emacsclient -n {0}"

    The ``{0}`` will be replaced by the filename.

    Note that this must be a *non waiting* command, i.e. which
    launches the editor on the specified file in a new window and then
    returns control to the command line without waiting for that new
    window to terminate.



  .. attribute:: docs_rsync_dest

    A Python template string which defines the rsync destination for
    publishing your projects documentation.
    Used by :cmd:`fab pub`.

    Example::

      env.docs_rsync_dest = 'luc@example.org:~/public_html/{prj}_{docs}'

    The ``{prj}`` in this template will be replaced by the internal
    name of this project, and ``{{docs}}`` by the name of the doctree
    (taken from :attr:`doc_trees`).

    For backward compatibility the following (deprecated) template is
    also still allowed::

      env.docs_rsync_dest = 'luc@example.org:~/public_html/%s'

    The ``%s`` in this template will be replaced by a name `xxx_yyy`,
    where `xxx` is the internal name of this project and `yyy` the
    name of the doctree (taken from :attr:`doc_trees`).


  .. attribute:: doc_trees

    A list of directory names (relative to your project directory)
    containing Sphinx document trees.
    Default value is ``['docs']``

    If this project has a main package, then `env.doc_trees` will be
    replaced by `doc_trees` attribute of that module.

  .. attribute:: cleanable_files

    A list of wildcards to be cleaned by :cmd:`fab clean`.

  .. attribute:: use_dirhtml

    Whether `sphinx-build
    <http://sphinx-doc.org/invocation.html#invocation-of-sphinx-build>`__
    should use ``dirhtml`` instead of the default ``html`` builder.

  .. attribute:: tolerate_sphinx_warnings

    Whether `sphinx-build` should tolerate warnings.

  .. attribute:: languages

    A list of language codes for which userdocs are being maintained.

  .. attribute:: apidoc_exclude_pathnames

    No longer used because we now use autosummary instead of
    sphinx-apidoc.

    a list of filenames (or directory names) to be excluded when you
    run :cmd:`fab api`.

  .. attribute:: revision_control_system

    The revision control system used by your project.
    Allowed values are `'git'`, `'hg'` or `None`.
    Used by :cmd:`fab ci`.

  .. attribute:: use_mercurial

    **No longer used.** Use :attr:`env.revision_control_system` instead.)

  .. attribute:: demo_projects

    The list of *Django demo projects* included in this project.

    Django demo projects are used by the test suite and the Sphinx
    documentation.  Before running :command:`fab test` or
    :command:`fab bd`, they must have been initialized.  To initialize
    them, run :command:`fab initdb`.

    It is not launched automatically by :command:`fab test` or
    :command:`fab bd` because it can take some time and is not always
    necessary.


History
=======

- 20141020 moved `doc_trees` project to :class:`atelier.Project`.
- 20141001 added support for multiple doc trees per project
  (:attr:`env.doc_trees`).
- 20140116 : added support for managing namespace packages

TODO
====

- replace `env.blogger_project` by an attribute of the main module
  (like `intersphinx_urls`)

(The rest of this page is automatically generated stuff.)

"""
import importlib
import os
import datetime
import glob

import sphinx
from babel.dates import format_date
from unipath import Path

from atelier.utils import i2d
from atelier import rstgen

try:
    from fabric.api import env, local, task
    from fabric.utils import abort, puts
    from fabric.contrib.console import confirm
    from fabric.api import lcd
except ImportError:
    def task(**kwargs):
        def d(f):
            return f
        return d
    # ignore it here so that Sphinx autodoc can import it even
    # if fabric is not installed.


def get_current_date(today=None):
    """
    """

    if today is None:
        return datetime.date.today()
    return i2d(today)


class RstFile(object):

    def __init__(self, local_root, url_root, parts):
        self.path = local_root.child(*parts) + '.rst'
        self.url = url_root + "/" + "/".join(parts) + '.html'
        # if parts[0] == 'docs':
        #     self.url = url_root + "/" + "/".join(parts[1:]) + '.html'
        # else:
        #     raise Exception("20131125")
            # self.url = url_root + "/" + "/".join(parts) + '.html'


def add_demo_project(p):
    """Register the specified settings module as being a Django demo project.
    See also :attr:`env.demo_projects`.

    """
    if p in env.demo_projects:
        return
        # raise Exception("Duplicate entry %r in demo_projects." % db)
    env.demo_projects.append(p)


def setup_from_fabfile(
        globals_dict, main_package=None, settings_module_name=None):
    """To be called from within your project's :xfile:`fabfile.py`.

    Minimal example::

      from atelier.fablib import *
      setup_from_fabfile(globals())

    If this doctree is the main doctree of a Python project, then the
    minimal example should be::

      from atelier.fablib import *
      setup_from_fabfile(globals(), "foobar")

    Where "foobar" is the Python name of your project's main package.

    """
    if not '__file__' in globals_dict:
        raise Exception(
            "No '__file__' in %r. "
            "First parameter to must be `globals()`" % globals_dict)
        
    fabfile = Path(globals_dict['__file__'])
    if not fabfile.exists():
        raise Exception("No such file: %s" % fabfile)
    env.root_dir = fabfile.parent.absolute()
    # print("20141027 %s %s " % (main_package, env.root_dir))

    env.project_name = env.root_dir.name
    env.setdefault('build_dir_name', '.build')  # but ablog needs '_build'
    
    env.setdefault('long_date_format', "%Y%m%d (%A, %d %B %Y)")
    # env.work_root = Path(env.work_root)
    env.setdefault('use_dirhtml', False)
    env.setdefault('blog_root', env.root_dir.child('docs'))

    env.setdefault('sdist_dir', None)
    env.setdefault('editor_command', None)
    if env.sdist_dir is not None:
        env.sdist_dir = Path(env.sdist_dir)
    env.main_package = main_package
    env.locale_dir = None
    env.tolerate_sphinx_warnings = False
    env.demo_projects = []
    env.revision_control_system = None
    env.apidoc_exclude_pathnames = []
    # env.blogger_url = "http://blog.example.com/"

    env.setdefault('languages', None)
    env.setdefault('blogger_project', None)
    env.setdefault('blogger_url', None)
    env.setdefault('cleanable_files', [])

    if isinstance(env.languages, basestring):
        env.languages = env.languages.split()

    # if env.main_package:
    #     env.SETUP_INFO = get_setup_info(Path(env.root_dir))
    # else:
    #     env.SETUP_INFO = None

    if settings_module_name is not None:
        os.environ['DJANGO_SETTINGS_MODULE'] = settings_module_name
        from django.conf import settings
        # why was this? settings.SITE.startup()
        env.languages = [lng.name for lng in settings.SITE.languages]
        # env.demo_databases.append(settings_module_name)
        #~ env.userdocs_base_language = settings.SITE.languages[0].name

    # The following import will populate the projects
    from atelier.projects import get_project_info
    env.current_project = get_project_info(env.root_dir)

    env.doc_trees = env.current_project.doc_trees

    # env.SETUP_INFO = env.current_project.SETUP_INFO


setup_from_project = setup_from_fabfile  # backwards compat


#~ def confirm(msg,default='y',others='n',**override_callbacks):
    #~ text = "%s [%s%s]" % (msg,default.upper(),others)
    #~ def y(): return True
    # ~ # def n(): abort("Missing user confirmation for:\n%s" % msg)
    #~ def n(): abort("Missing user confirmation")
    #~ callbacks = dict(y=y,n=n)
    #~ callbacks.update(override_callbacks)
    #~ while True:
        #~ answer = prompt(text)
        # ~ # answer = raw_input(prompt)
        #~ if not answer:
            #~ answer = default
        #~ answer = answer.lower()
        #~ if answer:
            #~ return callbacks.get(answer)()
def must_confirm(*args, **kw):
    if not confirm(*args, **kw):
        abort("Dann eben nicht...")


def must_exist(p):
    if not p.exists():
        abort("No such file: %s" % p.absolute())


def rmtree_after_confirm(p):
    if not p.exists():
        return
    if confirm("OK to remove %s and everything under it?" % p.absolute()):
        p.rmtree()


def unused_get_locale_dir():
    # replaced by env.locale_dir
    if not env.main_package:
        return None  # abort("No main_package")
    args = env.main_package.split('.')
    args.append('locale')
    p = env.root_dir.child(*args)
    if not p.isdir():
        return None  # abort("Directory %s does not exist." % p)
    return p


def cleanup_pyc(p):
    """Thanks to oddthinking on http://stackoverflow.com/questions/2528283
    """
    for root, dirs, files in os.walk(p):
        pyc_files = [filename for filename in files if filename.endswith(".pyc")]
        py_files = set([filename for filename in files if filename.endswith(".py")])
        excess_pyc_files = [pyc_filename for pyc_filename in pyc_files if pyc_filename[:-1] not in py_files]
        for excess_pyc_file in excess_pyc_files:
            full_path = os.path.join(root, excess_pyc_file)
            must_confirm("Remove excess file %s:" % full_path)
            os.remove(full_path)


@task(alias='unused_mm')
def make_messages():
    "Extract messages, then initialize and update all catalogs."
    extract_messages()
    init_catalog_code()
    update_catalog_code()

    if False:
        extract_messages_userdocs()
        setup_babel_userdocs('init_catalog')
        setup_babel_userdocs('update_catalog')


def extract_messages():
    """Extract messages from source files to `django.pot` file"""
    # locale_dir = get_locale_dir()
    locale_dir = env.locale_dir
    if locale_dir is None:
        return
    args = ["python", "setup.py"]
    args += ["extract_messages"]
    args += ["-o", Path(locale_dir).child("django.pot")]
    cmd = ' '.join(args)
    #~ must_confirm(cmd)
    local(cmd)


def extract_messages_userdocs():
    """
    Run the Sphinx gettext builder on userdocs.
    """
    userdocs = env.root_dir.child('userdocs')
    if not userdocs.isdir():
        return  # abort("Directory %s does not exist." % userdocs)
    args = ['sphinx-build', '-b', 'gettext']
    #~ args += cmdline_args
    # ~ args += ['-a'] # all files, not only outdated
    # ~ args += ['-P'] # no postmortem
    # ~ args += ['-Q'] # no output
    #~ if not env.tolerate_sphinx_warnings:
        # ~ args += ['-W'] # consider warnings as errors
    #~ args += ['-w',env.DOCSDIR.child('warnings.txt')]
    args += [userdocs]
    args += [userdocs.child("translations")]
    cmd = ' '.join(args)
    local(cmd)


@task(alias='rename')
def rename_data_url_friendly():
    data_dir = env.root_dir.child('docs', 'data')
    #~ print list(data_dir.listdir(names_only=True))
    print(list(data_dir.walk()))


def setup_babel_userdocs(babelcmd):
    """Create userdocs .po files if necessary."""
    userdocs = env.root_dir.child('userdocs')
    if not userdocs.isdir():
        return
    locale_dir = userdocs.child('translations')
    for domain in locale_dir.listdir('*.pot', names_only=True):
        domain = domain[:-4]
        for loc in env.languages:
            if loc != env.languages[0]:
                po_file = Path(locale_dir, loc, 'LC_MESSAGES', '%s.po' %
                               domain)
                mo_file = Path(locale_dir, loc, 'LC_MESSAGES', '%s.mo' %
                               domain)
                pot_file = Path(locale_dir, '%s.pot' % domain)
                if babelcmd == 'init_catalog' and po_file.exists():
                    print("Skip %s because file exists." % po_file)
                #~ elif babelcmd == 'compile_catalog' and not mo_file.needs_update(po_file):
                    #~ print "Skip %s because newer than .po" % mo_file
                else:
                    args = ["python", "setup.py"]
                    args += [babelcmd]
                    args += ["-l", loc]
                    args += ["--domain", domain]
                    args += ["-d", locale_dir]
                    #~ args += [ "-o" , po_file ]
                    #~ if babelcmd == 'init_catalog':
                    if babelcmd == 'compile_catalog':
                        args += ["-i", po_file]
                    else:
                        args += ["-i", pot_file]
                    cmd = ' '.join(args)
                    #~ must_confirm(cmd)
                    local(cmd)


@task(alias='cmu')
def compile_catalog_userdocs():
    setup_babel_userdocs('compile_catalog')


def init_catalog_code():
    """Create code .po files if necessary."""
    from lino.core.site import to_locale
    locale_dir = env.locale_dir
    # locale_dir = get_locale_dir()
    if locale_dir is None:
        return
    locale_dir = Path(locale_dir)
    for loc in env.languages:
        if loc != 'en':
            f = locale_dir.child(loc, 'LC_MESSAGES', 'django.po')
            if f.exists():
                print("Skip %s because file exists." % f)
            else:
                args = ["python", "setup.py"]
                args += ["init_catalog"]
                args += ["--domain django"]
                args += ["-l", to_locale(loc)]
                args += ["-d", locale_dir]
                #~ args += [ "-o" , f ]
                args += ["-i", locale_dir.child('django.pot')]
                cmd = ' '.join(args)
                must_confirm(cmd)
                local(cmd)


def update_catalog_code():
    """Update .po files from .pot file."""
    from lino.core.site import to_locale
    locale_dir = env.locale_dir
    # locale_dir = get_locale_dir()
    if locale_dir is None:
        return
    locale_dir = Path(locale_dir)
    for loc in env.languages:
        if loc != env.languages[0]:
            args = ["python", "setup.py"]
            args += ["update_catalog"]
            args += ["--domain django"]
            #~ args += [ "-d" , locale_dir ]
            args += ["-o", locale_dir.child(loc, 'LC_MESSAGES', 'django.po')]
            args += ["-i", locale_dir.child("django.pot")]
            args += ["-l", to_locale(loc)]
            cmd = ' '.join(args)
            #~ must_confirm(cmd)
            local(cmd)


@task(alias='cm')
def compile_catalog():
    """Compile .po files to .mo files."""
    from lino.core.site import to_locale 
    locale_dir = env.locale_dir
    # locale_dir = get_locale_dir()
    if locale_dir is None:
        return
    for loc in env.languages:
        if loc != env.languages[0]:
            args = ["python", "setup.py"]
            args += ["compile_catalog"]
            args += ["-i", locale_dir.child(loc, 'LC_MESSAGES', 'django.po')]
            args += ["-o", locale_dir.child(loc, 'LC_MESSAGES', 'django.mo')]
            args += ["--domain django"]
            #~ args += [ "-d" , locale_dir ]
            args += ["-l", to_locale(loc)]
            cmd = ' '.join(args)
            #~ must_confirm(cmd)
            local(cmd)


@task(alias='mss')
def makescreenshots():
    """generate screenshot .jpg files to gen/screenshots."""
    run_in_demo_projects('makescreenshots', '--traceback')


@task(alias='sss')
def syncscreenshots():
    """synchronize gen/screenshots to userdocs/gen/screenshots."""
    run_in_demo_projects('syncscreenshots', '--traceback',
                          'gen/screenshots', 'userdocs/gen/screenshots')


def sphinx_build(builder, docs_dir,
                 cmdline_args=[], language=None, build_dir_cmd=None):
    args = ['sphinx-build', '-b', builder]
    args += cmdline_args
    # ~ args += ['-a'] # all files, not only outdated
    # ~ args += ['-P'] # no postmortem
    # ~ args += ['-Q'] # no output
    # build_dir = docs_dir.child(env.build_dir_name)
    build_dir = Path(env.build_dir_name)
    if language is not None:
        args += ['-D', 'language=' + language]
        # needed in select_lang.html template
        args += ['-A', 'language=' + language]
        if language != env.languages[0]:
            build_dir = build_dir.child(language)
            #~ print 20130726, build_dir
    if env.tolerate_sphinx_warnings:
        args += ['-w', 'warnings_%s.txt' % builder]
    else:
        args += ['-W']  # consider warnings as errors
        # args += ['-vvv']  # increase verbosity
    #~ args += ['-w'+Path(env.root_dir,'sphinx_doctest_warnings.txt')]
    args += ['.', build_dir]
    cmd = ' '.join(args)
    with lcd(docs_dir):
        local(cmd)
    if build_dir_cmd is not None:
        with lcd(build_dir):
            local(build_dir_cmd)


def sync_docs_data(docs_dir):
    build_dir = docs_dir.child(env.build_dir_name)
    for data in ('dl', 'data'):
        src = docs_dir.child(data).absolute()
        if src.isdir():
            target = build_dir.child('dl')
            target.mkdir()
            cmd = 'cp -ur %s %s' % (src, target.parent)
            local(cmd)
    if False:
        # according to http://mathiasbynens.be/notes/rel-shortcut-icon
        for n in ['favicon.ico']:
            src = docs_dir.child(n).absolute()
            if src.exists():
                target = build_dir.child(n)
                cmd = 'cp %s %s' % (src, target.parent)
                local(cmd)


@task(alias='userdocs')
def build_userdocs(*cmdline_args):
    """
    Deprecated. sphinx-build the userdocs tree in all languages
    """
    if env.languages is None:
        return
    docs_dir = env.root_dir.child('userdocs')
    if not docs_dir.exists():
        return
    for lng in env.languages:
        sphinx_build('html', docs_dir, cmdline_args, lng)
    sync_docs_data(docs_dir)


@task(alias='pdf')
def build_userdocs_pdf(*cmdline_args):
    if env.languages is None:
        return
    docs_dir = env.root_dir.child('userdocs')
    if not docs_dir.exists():
        return
    for lng in env.languages:
        sphinx_build('latex', docs_dir, cmdline_args,
                     lng, build_dir_cmd='make all-pdf')
    sync_docs_data(docs_dir)


@task(alias='linkcheck')
def sphinx_build_linkcheck(*cmdline_args):
    """sphinxbuild -b linkcheck docs."""
    docs_dir = env.root_dir.child('docs')
    if docs_dir.exists():
        sphinx_build('linkcheck', docs_dir, cmdline_args)
    docs_dir = env.root_dir.child('userdocs')
    if docs_dir.exists():
        lng = env.languages[0]
        #~ lng = env.userdocs_base_language
        sphinx_build('linkcheck', docs_dir, cmdline_args, lng)


def get_doc_trees():
    for rel_doc_tree in env.doc_trees:
        docs_dir = env.root_dir.child(rel_doc_tree)
        if not docs_dir.exists():
            msg = "Directory %s does not exist." % docs_dir
            msg += "\nCheck your project's `doc_trees` setting."
            raise Exception(msg)
        yield docs_dir


def run_in_demo_projects(admin_cmd, *more):
    """Run the given shell command in each demo project (see
    :attr:`env.demo_projects`).

    """
    for mod in env.demo_projects:
        puts("-" * 80)
        puts("In demo project {0}:".format(mod))

        from importlib import import_module
        m = import_module(mod)
        # p = Path(m.__file__).parent.absolute()
        p = m.SITE.cache_dir or m.SITE.project_dir

        with lcd(p):
            args = ["django-admin.py"]
            args += [admin_cmd]
            args += more
            #~ args += ["--noinput"]
            args += ["--settings=" + mod]
            #~ args += [" --pythonpath=%s" % p.absolute()]
            cmd = " ".join(args)
            local(cmd)





@task(alias='ddt')
def double_dump_test():
    """
    Perform a "double dump test" on every demo database.
    TODO: convert this to a Lino management command.
    """
    raise Exception("Not yet converted after 20150129")
    if len(env.demo_databases) == 0:
        return
    a = Path(env.temp_dir, 'a')
    b = Path(env.temp_dir, 'b')
    rmtree_after_confirm(a)
    rmtree_after_confirm(b)
    #~ if not confirm("This will possibly break the demo databases. Are you sure?"):
        #~ return
    #~ a.mkdir()
    with lcd(env.temp_dir):
        for db in env.demo_databases:
            if a.exists():
                a.rmtree()
            if b.exists():
                b.rmtree()
            local("django-admin.py dump2py --settings=%s --traceback a" % db)
            local(
                "django-admin.py run --settings=%s --traceback a/restore.py" %
                db)
            local("django-admin.py dump2py --settings=%s --traceback b" % db)
            local("diff a b")





# -*- coding: utf-8 -*-
# Copyright 2011-2017 by Luc Saffre.
# License: BSD, see LICENSE for more details.

"""Defines the :class:`InsertInputDirective` class and some
subclasses, installing the :rst:dir:`py2rst` directive.

To use it, you must run::

  $ pip install atelier

and then add this module to your :xfile:`conf.py`::

  extensions += ['atelier.sphinxconf.insert_input']

.. rst:directive:: py2rst

Executes the Python code, capturing the `stdout` and inserting it to
the document, parsing it as reStructuredText.

For example, if you write::

  .. py2rst::

      url = 'http://planet.python.org/'
      print("`This <%s>`_ is my *favourite* planet." % url)

then you get:

.. py2rst::

  url = 'http://planet.python.org/'
  print("`This <%s>`_ is my *favourite* planet." % url)

**Warning**: installing this extension makes your Sphinx instance
unsecure. That is, you should not use this in an environment where
arbitrary content can be posted, since that content is actually being
executed with the permissions of the process that runs the Sphinx
builder.

Note that when the Sphinx builder is running under Python 2.7, the
following future imports have been done::

  from __future__ import print_function
  from __future__ import unicode_literals

"""

from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from future.utils import PY3

"""Note: the `import unicode_literals` caused the following::

    Traceback (most recent call last):
      File "/home/luc/pythonenvs/py27/local/lib/python2.7/site-packages/sphinx/cmdline.py", line 245, in main
        warningiserror, tags, verbosity, parallel)
      File "/home/luc/pythonenvs/py27/local/lib/python2.7/site-packages/sphinx/application.py", line 122, in __init__
        self.config.setup(self)
      File "/home/luc/hgwork/atelier/atelier/sphinxconf/__init__.py", line 654, in setup
        indextemplate='pair: %s; management command')
      File "/home/luc/pythonenvs/py27/local/lib/python2.7/site-packages/sphinx/application.py", line 503, in add_object_type
        'doc_field_types': doc_field_types})
    TypeError: type() argument 1 must be string, not unicode
    
I solved this by a manual patch in line 308 of
:file:`sphinx/application.py`::
    
    def import_object(self, objname, source=None):
        objname = str(objname)  # LS 20140108 accept unicode strings
        # needed when calling from Python 2.7 with
        # `from __future__ import unicode_literals`
        try:
            module, name = objname.rsplit('.', 1)
        except ValueError, err:
    

"""


import logging
logger = logging.getLogger(__name__)

import sys
if PY3:
    from io import BytesIO as StringIO  # see blog 20160125, 20160218
else:
    from StringIO import StringIO

import traceback

from docutils import nodes
from docutils import statemachine
from docutils.parsers.rst import directives
from docutils.parsers.rst import Directive
#from sphinx.util.compat import Directive
from sphinx.util.nodes import nested_parse_with_titles

#~ class ScreenshotDirective(directives.images.Image):
    #~ """
    #~ Directive to insert a screenshot.
    #~ """
    #~ def run(self):
        #~ assert len(self.arguments) == 1
        # ~ # name = '/../gen/screenshots/' + self.arguments[0]
        #~ name = '/gen/screenshots/' + self.arguments[0]
        #~ self.arguments = [name]
        #~ (image_node,) = directives.images.Image.run(self)
        #~ return [image_node]


class InsertInputDirective(Directive):
    """
    Base class for directives that work by generating rst markup
    to be forwarded to `state_machine.insert_input()`.
    """
    titles_allowed = False
    has_content = True
    raw_insert = False
    option_spec = {
        'language': directives.unchanged_required,
        'debug': directives.flag,
    }

    def get_rst(self):
        "Override this to return a string in rst syntax"
        raise NotImplementedError()

    def run(self):

        self.env = self.state.document.settings.env
        self.language = self.options.get('language', self.env.config.language)
        self.env.temp_data['language'] = self.language

        # catch exceptions and report them together with the name of
        # the guilty file
        try:
            output = self.get_rst()
        except Exception as e:
            traceback.print_exc(e)
            document = self.state.document
            return [document.reporter.warning(str(e), line=self.lineno)]

        #~ output = output.decode('utf-8')

        if 'debug' in self.options:
            print(self.env.docname)
            print('-' * 50)
            print(output)
            print('-' * 50)

        content = statemachine.StringList(output.splitlines())

        if self.raw_insert:

            self.state_machine.insert_input(content, output)
            return []

        if self.titles_allowed:
            node = nodes.section()
            # necessary so that the child nodes get the right source/line set
            node.document = self.state.document
            nested_parse_with_titles(self.state, content, node)
        else:
            node = nodes.paragraph()
            node.document = self.state.document
            self.state.nested_parse(content, self.content_offset, node)

        # following lines originally copied from
        # docutils.parsers.rst.directives.tables.RSTTable
        #~ title, messages = self.make_title()
        # ~ node = nodes.Element()          # anonymous container for parsing
        #~ self.state.nested_parse(content, self.content_offset, node)
        #~ if len(node) != 1 or not isinstance(node[0], nodes.table):
            #~ error = self.state_machine.reporter.error(
                #~ 'Error parsing content block for the "%s" directive: exactly '
                #~ 'one table expected.' % self.name, nodes.literal_block(
                #~ self.block_text, self.block_text), line=self.lineno)
            #~ return [error]
        #~ return [x for x in node]
        return list(node)

        #~ table_node = node[0]
        #~ table_node['classes'] += self.options.get('class', [])
        #~ return [table_node]


class Py2rstDirective(InsertInputDirective):

    """Implements the :rst:dir:`py2rst` directive."""

    titles_allowed = True
    has_content = True

    def get_context(self):
        context = dict()
        context.update(self=self)
        context.update(env=self.state.document.settings.env)
        return context

    def get_rst(self):
        if not self.content:
            warning = self.state_machine.reporter.warning(
                'Content block expected for the "%s" directive; none found.'
                % self.name, nodes.literal_block(
                    self.block_text, self.block_text), line=self.lineno)
            return [warning]

        #~ raise Exception("20130331 %r" % self.content)
        code = '\n'.join(self.content)
        return self.output_from_exec(code)

    def output_from_exec(self, code):
        old = sys.stdout
        buffer = StringIO()
        sys.stdout = buffer
        context = self.get_context()

        if 'debug' in self.options:
            exec(code, context)
        else:
            try:
                exec(code, context)
            except Exception as err:
                # f = inspect.trace()[1]
                # traceback.print_stack()
                # code = code.replace("%", "\%")
                # raise Exception("%s in code:\n%s" % (err, code))
                raise Exception("{} in code:\n{}".format(err, code))

        sys.stdout = old
        s = buffer.getvalue()
        #~ print 20130331, type(s)
        return s

    def shell_block(self, cmd):
        """Run the given command and insert a :rst:dir:`code-block` directive
        which displays both the command and its output.

        For example, if your `.rst` document contains::

            .. py2rst::

                self.shell_block(["echo", "Hello", "world!"])
            
        Then it will be rendered as:

        .. py2rst::

            self.shell_block(["echo", "Hello", "world!"])
        
        This uses the `subprocess.check_output
        <https://docs.python.org/2/library/subprocess.html#subprocess.check_output>`_
        method and the security warnings apply.
        
        If the command returns with a non-zero exit code, the
        exception is catched and converted into a warning.

        """
        print(".. code-block:: bash")
        print()
        import subprocess
        print("    $ " + ' '.join(cmd))
        for ln in subprocess.check_output(cmd).splitlines():
            print("    " + ln)


#~ class DjangoTableDirective(InsertInputDirective):
    #~ def get_rst(self):
        #~ assert len(self.content) == 1
        #~ code = '\n'.join(self.content)
        #~ from django.conf import settings
        #~ print .jobs.Candidatures.request(limit=5).to_rst()
        #~ code = """
        #~ """
        #~ old = sys.stdout
        #~ buffer = StringIO()
        #~ sys.stdout = buffer
        #~ context = dict()
        #~ context.update(settings.SITE.modules)
        #~ context = dict(settings=settings)
        #~ exec(code,context)
        #~ sys.stdout = old
        #~ return buffer.getvalue()


# class BlogNoteDirective(Py2rstDirective):

#     def get_rst(self):
#         return '\n'.join(self.content)


def setup(app):
    # also used by `vor/conf.py`
    app.add_directive(str('py2rst'), Py2rstDirective)




# -*- coding: UTF-8 -*-
# Copyright 2016-2017 by Luc Saffre & Hamza Khchine.
# License: BSD, see LICENSE for more details.
"""A little hack for Lino developers who also have non-Lino projects.

Because otherwise commands like :cmd:`pp inv prep` would fail on
non-Lino projects because they wouldn't know what :cmd:`inv prep`
means.

"""

from .ns import ns

# from invoke import task

# @task(name='prep')
# def prep(ctx, *args, **kwargs):
#     """Does nothing. """

# ns.add_task(prep)
